@@ -2,11 +2,11 @@ import atexit
 import asyncio
 import datetime
 import importlib.util
+import json
 import os
 import signal
-import sys
 import threading
-import types
+import sys
 from types import FrameType
 from typing import Any, Iterable
 
@@ -20,8 +20,9 @@ from rich.table import Table
 from . import flow, lib, setting
 from .setup import flow_names_with_setup
 from .runtime import execution_context
+from .subprocess_exec import add_user_app
+from .user_app_loader import load_user_app, Error as UserAppLoaderError
 
-# Create ServerSettings lazily upon first call, as environment variables may be loaded from files, etc.
 COCOINDEX_HOST = "https://cocoindex.io"
 
 
@@ -76,55 +77,19 @@ def _get_app_ref_from_specifier(
     return app_ref
 
 
-def _load_user_app(app_target: str) -> types.ModuleType:
-    """
-    Loads the user's application, which can be a file path or an installed module name.
-    Exits on failure.
-    """
+def _load_user_app(app_target: str) -> None:
     if not app_target:
         raise click.ClickException("Application target not provided.")
 
-    looks_like_path = os.sep in app_target or app_target.lower().endswith(".py")
-
-    if looks_like_path:
-        if not os.path.isfile(app_target):
-            raise click.ClickException(f"Application file path not found: {app_target}")
-        app_path = os.path.abspath(app_target)
-        app_dir = os.path.dirname(app_path)
-        module_name = os.path.splitext(os.path.basename(app_path))[0]
-
-        if app_dir not in sys.path:
-            sys.path.insert(0, app_dir)
-        try:
-            spec = importlib.util.spec_from_file_location(module_name, app_path)
-            if spec is None:
-                raise ImportError(f"Could not create spec for file: {app_path}")
-            module = importlib.util.module_from_spec(spec)
-            sys.modules[spec.name] = module
-            if spec.loader is None:
-                raise ImportError(f"Could not create loader for file: {app_path}")
-            spec.loader.exec_module(module)
-            return module
-        except (ImportError, FileNotFoundError, PermissionError) as e:
-            raise click.ClickException(f"Failed importing file '{app_path}': {e}")
-        finally:
-            if app_dir in sys.path and sys.path[0] == app_dir:
-                sys.path.pop(0)
-
-    # Try as module
     try:
-        return importlib.import_module(app_target)
-    except ImportError as e:
-        raise click.ClickException(f"Failed to load module '{app_target}': {e}")
-    except Exception as e:
-        raise click.ClickException(
-            f"Unexpected error importing module '{app_target}': {e}"
-        )
+        load_user_app(app_target)
+    except UserAppLoaderError as e:
+        raise ValueError(f"Failed to load APP_TARGET '{app_target}'") from e
+
+    add_user_app(app_target)
 
 
 def _initialize_cocoindex_in_process() -> None:
-    settings = setting.Settings.from_env()
-    lib.init(settings)
     atexit.register(lib.stop)
 
 
@@ -140,7 +105,13 @@ def _initialize_cocoindex_in_process() -> None:
     default=None,
     show_default=False,
 )
-def cli(env_file: str | None = None) -> None:
+@click.option(
+    "--app-dir",
+    help="Load apps from the specified directory. Default to the current directory.",
+    default="",
+    show_default=True,
+)
+def cli(env_file: str | None = None, app_dir: str | None = "") -> None:
     """
     CLI for Cocoindex.
     """
@@ -150,6 +121,9 @@ def cli(env_file: str | None = None) -> None:
         loaded_env_path = os.path.abspath(dotenv_path)
         click.echo(f"Loaded environment variables from: {loaded_env_path}\n", err=True)
 
+    if app_dir is not None:
+        sys.path.insert(0, app_dir)
+
     try:
         _initialize_cocoindex_in_process()
     except Exception as e:
@@ -162,11 +136,9 @@ def ls(app_target: str | None) -> None:
     """
     List all flows.
 
-    If APP_TARGET (path/to/app.py or a module) is provided, lists flows
-    defined in the app and their backend setup status.
+    If `APP_TARGET` (`path/to/app.py` or a module) is provided, lists flows defined in the app and their backend setup status.
 
-    If APP_TARGET is omitted, lists all flows that have a persisted
-    setup in the backend.
+    If `APP_TARGET` is omitted, lists all flows that have a persisted setup in the backend.
     """
     persisted_flow_names = flow_names_with_setup()
     if app_target:
@@ -214,16 +186,15 @@ def show(app_flow_specifier: str, color: bool, verbose: bool) -> None:
     """
     Show the flow spec and schema.
 
-    APP_FLOW_SPECIFIER: Specifies the application and optionally the target flow.
-    Can be one of the following formats:
+    `APP_FLOW_SPECIFIER`: Specifies the application and optionally the target flow. Can be one of the following formats:
 
     \b
-      - path/to/your_app.py
-      - an_installed.module_name
-      - path/to/your_app.py:SpecificFlowName
-      - an_installed.module_name:SpecificFlowName
+      - `path/to/your_app.py`
+      - `an_installed.module_name`
+      - `path/to/your_app.py:SpecificFlowName`
+      - `an_installed.module_name:SpecificFlowName`
 
-    :SpecificFlowName can be omitted only if the application defines a single flow.
+    `:SpecificFlowName` can be omitted only if the application defines a single flow.
     """
     app_ref, flow_ref = _parse_app_flow_specifier(app_flow_specifier)
     _load_user_app(app_ref)
@@ -245,6 +216,58 @@ def show(app_flow_specifier: str, color: bool, verbose: bool) -> None:
     console.print(table)
 
 
+def _drop_flows(flows: Iterable[flow.Flow], app_ref: str, force: bool = False) -> None:
+    """
+    Helper function to drop flows without user interaction.
+    Used internally by --reset flag
+
+    Args:
+        flows: Iterable of Flow objects to drop
+        force: If True, skip confirmation prompts
+    """
+    flow_full_names = ", ".join(fl.full_name for fl in flows)
+    click.echo(
+        f"Preparing to drop specified flows: {flow_full_names} (in '{app_ref}').",
+        err=True,
+    )
+
+    if not flows:
+        click.echo("No flows identified for the drop operation.")
+        return
+
+    setup_bundle = flow.make_drop_bundle(flows)
+    description, is_up_to_date = setup_bundle.describe()
+    click.echo(description)
+    if is_up_to_date:
+        click.echo("No flows need to be dropped.")
+        return
+    if not force and not click.confirm(
+        f"\nThis will apply changes to drop setup for: {flow_full_names}. Continue? [yes/N]",
+        default=False,
+        show_default=False,
+    ):
+        click.echo("Drop operation aborted by user.")
+        return
+    setup_bundle.apply(report_to_stdout=True)
+
+
+def _deprecate_setup_flag(
+    ctx: click.Context, param: click.Parameter, value: bool
+) -> bool:
+    """Callback to warn users that --setup flag is deprecated."""
+    # Check if the parameter was explicitly provided by the user
+    if param.name is not None:
+        param_source = ctx.get_parameter_source(param.name)
+        if param_source == click.core.ParameterSource.COMMANDLINE:
+            click.secho(
+                "Warning: The --setup flag is deprecated and will be removed in a future version. "
+                "Setup is now always enabled by default.",
+                fg="yellow",
+                err=True,
+            )
+    return value
+
+
 def _setup_flows(
     flow_iter: Iterable[flow.Flow],
     *,
@@ -294,14 +317,26 @@ async def _update_all_flows_with_hint_async(
     default=False,
     help="Force setup without confirmation prompts.",
 )
-def setup(app_target: str, force: bool) -> None:
+@click.option(
+    "--reset",
+    is_flag=True,
+    show_default=True,
+    default=False,
+    help="Drop existing setup before running setup (equivalent to running 'cocoindex drop' first).",
+)
+def setup(app_target: str, force: bool, reset: bool) -> None:
     """
     Check and apply backend setup changes for flows, including the internal storage and target (to export to).
 
-    APP_TARGET: path/to/app.py or installed_module.
+    `APP_TARGET`: `path/to/app.py` or `installed_module`.
     """
     app_ref = _get_app_ref_from_specifier(app_target)
     _load_user_app(app_ref)
+
+    # If --reset is specified, drop existing setup first
+    if reset:
+        _drop_flows(flow.flows().values(), app_ref=app_ref, force=force)
+
     _setup_flows(flow.flows().values(), force=force, always_show_setup=True)
 
 
@@ -350,30 +385,7 @@ def drop(app_target: str | None, flow_name: tuple[str, ...], force: bool) -> Non
     else:
         flows = flow.flows().values()
 
-    flow_full_names = ", ".join(fl.full_name for fl in flows)
-    click.echo(
-        f"Preparing to drop specified flows: {flow_full_names} (in '{app_ref}').",
-        err=True,
-    )
-
-    if not flows:
-        click.echo("No flows identified for the drop operation.")
-        return
-
-    setup_bundle = flow.make_drop_bundle(flows)
-    description, is_up_to_date = setup_bundle.describe()
-    click.echo(description)
-    if is_up_to_date:
-        click.echo("No flows need to be dropped.")
-        return
-    if not force and not click.confirm(
-        f"\nThis will apply changes to drop setup for: {flow_full_names}. Continue? [yes/N]",
-        default=False,
-        show_default=False,
-    ):
-        click.echo("Drop operation aborted by user.")
-        return
-    setup_bundle.apply(report_to_stdout=True)
+    _drop_flows(flows, app_ref=app_ref, force=force)
 
 
 @cli.command()
@@ -387,11 +399,26 @@ def drop(app_target: str | None, flow_name: tuple[str, ...], force: bool) -> Non
     help="Continuously watch changes from data sources and apply to the target index.",
 )
 @click.option(
-    "--setup",
+    "--reexport",
     is_flag=True,
     show_default=True,
     default=False,
-    help="Automatically setup backends for the flow if it's not setup yet.",
+    help="Reexport to targets even if there's no change.",
+)
+@click.option(
+    "--setup",
+    is_flag=True,
+    show_default=True,
+    default=True,
+    callback=_deprecate_setup_flag,
+    help="(DEPRECATED) Automatically setup backends for the flow if it's not setup yet. This is now the default behavior.",
+)
+@click.option(
+    "--reset",
+    is_flag=True,
+    show_default=True,
+    default=False,
+    help="Drop existing setup before updating (equivalent to running 'cocoindex drop' first). `--reset` implies `--setup`.",
 )
 @click.option(
     "-f",
@@ -412,18 +439,26 @@ def drop(app_target: str | None, flow_name: tuple[str, ...], force: bool) -> Non
 def update(
     app_flow_specifier: str,
     live: bool,
+    reexport: bool,
     setup: bool,  # pylint: disable=redefined-outer-name
+    reset: bool,
     force: bool,
     quiet: bool,
 ) -> None:
     """
     Update the index to reflect the latest data from data sources.
 
-    APP_FLOW_SPECIFIER: path/to/app.py, module, path/to/app.py:FlowName, or module:FlowName.
-    If :FlowName is omitted, updates all flows.
+    `APP_FLOW_SPECIFIER`: `path/to/app.py`, module, `path/to/app.py:FlowName`, or `module:FlowName`. If `:FlowName` is omitted, updates all flows.
     """
     app_ref, flow_name = _parse_app_flow_specifier(app_flow_specifier)
     _load_user_app(app_ref)
+    flow_list = (
+        [flow.flow_by_name(flow_name)] if flow_name else list(flow.flows().values())
+    )
+
+    # If --reset is specified, drop existing setup first
+    if reset:
+        _drop_flows(flow_list, app_ref=app_ref, force=force)
 
     if live:
         click.secho(
@@ -431,20 +466,19 @@ def update(
             fg="yellow",
         )
 
-    options = flow.FlowLiveUpdaterOptions(live_mode=live, print_stats=not quiet)
+    options = flow.FlowLiveUpdaterOptions(
+        live_mode=live,
+        reexport_targets=reexport,
+        print_stats=not quiet,
+    )
+    if reset or setup:
+        _setup_flows(flow_list, force=force, quiet=quiet)
+
     if flow_name is None:
-        if setup:
-            _setup_flows(
-                flow.flows().values(),
-                force=force,
-                quiet=quiet,
-            )
         execution_context.run(_update_all_flows_with_hint_async(options))
     else:
-        fl = flow.flow_by_name(flow_name)
-        if setup:
-            _setup_flows((fl,), force=force, quiet=quiet)
-        with flow.FlowLiveUpdater(fl, options) as updater:
+        assert len(flow_list) == 1
+        with flow.FlowLiveUpdater(flow_list[0], options) as updater:
             updater.wait()
             if options.live_mode:
                 _show_no_live_update_hint()
@@ -472,18 +506,16 @@ def evaluate(
     """
     Evaluate the flow and dump flow outputs to files.
 
-    Instead of updating the index, it dumps what should be indexed to files.
-    Mainly used for evaluation purpose.
+    Instead of updating the index, it dumps what should be indexed to files. Mainly used for evaluation purpose.
 
     \b
-    APP_FLOW_SPECIFIER: Specifies the application and optionally the target flow.
-    Can be one of the following formats:
-      - path/to/your_app.py
-      - an_installed.module_name
-      - path/to/your_app.py:SpecificFlowName
-      - an_installed.module_name:SpecificFlowName
+    `APP_FLOW_SPECIFIER`: Specifies the application and optionally the target flow. Can be one of the following formats:
+      - `path/to/your_app.py`
+      - `an_installed.module_name`
+      - `path/to/your_app.py:SpecificFlowName`
+      - `an_installed.module_name:SpecificFlowName`
 
-    :SpecificFlowName can be omitted only if the application defines a single flow.
+    `:SpecificFlowName` can be omitted only if the application defines a single flow.
     """
     app_ref, flow_ref = _parse_app_flow_specifier(app_flow_specifier)
     _load_user_app(app_ref)
@@ -539,8 +571,23 @@ def evaluate(
     "--setup",
     is_flag=True,
     show_default=True,
+    default=True,
+    callback=_deprecate_setup_flag,
+    help="(DEPRECATED) Automatically setup backends for the flow if it's not setup yet. This is now the default behavior.",
+)
+@click.option(
+    "--reset",
+    is_flag=True,
+    show_default=True,
     default=False,
-    help="Automatically setup backends for the flow if it's not setup yet.",
+    help="Drop existing setup before starting server (equivalent to running 'cocoindex drop' first). `--reset` implies `--setup`.",
+)
+@click.option(
+    "--reexport",
+    is_flag=True,
+    show_default=True,
+    default=False,
+    help="Reexport to targets even if there's no change.",
 )
 @click.option(
     "-f",
@@ -571,6 +618,8 @@ def server(
     address: str | None,
     live_update: bool,
     setup: bool,  # pylint: disable=redefined-outer-name
+    reset: bool,
+    reexport: bool,
     force: bool,
     quiet: bool,
     cors_origin: str | None,
@@ -583,7 +632,7 @@ def server(
 
     It will allow tools like CocoInsight to access the server.
 
-    APP_TARGET: path/to/app.py or installed_module.
+    `APP_TARGET`: `path/to/app.py` or `installed_module`.
     """
     app_ref = _get_app_ref_from_specifier(app_target)
     args = (
@@ -593,10 +642,14 @@ def server(
         cors_cocoindex,
         cors_local,
         live_update,
-        setup,
-        force,
+        reexport,
         quiet,
     )
+    kwargs = {
+        "run_reset": reset,
+        "run_setup": setup,
+        "force": force,
+    }
 
     if reload:
         watch_paths = {os.getcwd()}
@@ -614,6 +667,7 @@ def server(
             *watch_paths,
             target=_reloadable_server_target,
             args=args,
+            kwargs=kwargs,
             watch_filter=watchfiles.PythonFilter(),
             callback=lambda changes: click.secho(
                 f"\nDetected changes in {len(changes)} file(s), reloading server...\n",
@@ -625,12 +679,19 @@ def server(
             "NOTE: Flow code changes will NOT be reflected until you restart to load the new code. Use --reload to enable auto-reload.\n",
             fg="yellow",
         )
-        _run_server(*args)
+        _run_server(*args, **kwargs)
 
 
 def _reloadable_server_target(*args: Any, **kwargs: Any) -> None:
     """Reloadable target for the watchfiles process."""
     _initialize_cocoindex_in_process()
+
+    kwargs["run_setup"] = kwargs["run_setup"] or kwargs["run_reset"]
+    changed_files = json.loads(os.environ.get("WATCHFILES_CHANGES", "[]"))
+    if changed_files:
+        kwargs["run_reset"] = False
+    kwargs["force"] = True
+
     _run_server(*args, **kwargs)
 
 
@@ -641,13 +702,40 @@ def _run_server(
     cors_cocoindex: bool = False,
     cors_local: int | None = None,
     live_update: bool = False,
-    run_setup: bool = False,
-    force: bool = False,
+    reexport: bool = False,
     quiet: bool = False,
+    /,
+    *,
+    force: bool = False,
+    run_reset: bool = False,
+    run_setup: bool = False,
 ) -> None:
     """Helper function to run the server with specified settings."""
     _load_user_app(app_ref)
 
+    # Check if any flows are registered
+    if not flow.flow_names():
+        click.secho(
+            f"\nError: No flows registered in '{app_ref}'.\n",
+            fg="red",
+            bold=True,
+            err=True,
+        )
+        click.secho(
+            "To use CocoIndex server, you need to define at least one flow.",
+            err=True,
+        )
+        click.secho(
+            "See https://cocoindex.io/docs for more information.\n",
+            fg="cyan",
+            err=True,
+        )
+        raise click.Abort()
+
+    # If --reset is specified, drop existing setup first
+    if run_reset:
+        _drop_flows(flow.flows().values(), app_ref=app_ref, force=force)
+
     server_settings = setting.ServerSettings.from_env()
     cors_origins: set[str] = set(server_settings.cors_origins or [])
     if cors_origin is not None:
@@ -661,7 +749,7 @@ def _run_server(
     if address is not None:
         server_settings.address = address
 
-    if run_setup:
+    if run_reset or run_setup:
         _setup_flows(
             flow.flows().values(),
             force=force,
@@ -675,8 +763,12 @@ def _run_server(
 
     click.secho("Press Ctrl+C to stop the server.", fg="yellow")
 
-    if live_update:
-        options = flow.FlowLiveUpdaterOptions(live_mode=True, print_stats=not quiet)
+    if live_update or reexport:
+        options = flow.FlowLiveUpdaterOptions(
+            live_mode=live_update,
+            reexport_targets=reexport,
+            print_stats=not quiet,
+        )
         asyncio.run_coroutine_threadsafe(
             _update_all_flows_with_hint_async(options), execution_context.event_loop
         )
