@@ -9,15 +9,6 @@ import datetime
 import functools
 import inspect
 import re
-
-from .validation import (
-    validate_flow_name,
-    NamingError,
-    validate_full_flow_name,
-    validate_target_name,
-)
-from .typing import analyze_type_info
-
 from dataclasses import dataclass
 from enum import Enum
 from threading import Lock
@@ -25,13 +16,12 @@ from typing import (
     Any,
     Callable,
     Generic,
-    NamedTuple,
+    Iterable,
     Sequence,
     TypeVar,
     cast,
     get_args,
     get_origin,
-    Iterable,
 )
 
 from rich.text import Text
@@ -41,11 +31,21 @@ from . import _engine  # type: ignore
 from . import index
 from . import op
 from . import setting
-from .convert import dump_engine_object, encode_engine_value, make_engine_value_decoder
+from .engine_object import dump_engine_object
+from .engine_value import (
+    make_engine_value_decoder,
+    make_engine_value_encoder,
+)
 from .op import FunctionSpec
-from .runtime import execution_context
+from .runtime import execution_context, to_async_call
 from .setup import SetupChangeBundle
-from .typing import encode_enriched_type
+from .typing import analyze_type_info, encode_enriched_type, decode_engine_value_type
+from .query_handler import QueryHandlerInfo, QueryHandlerResultFields
+from .validation import (
+    validate_flow_name,
+    validate_full_flow_name,
+    validate_target_name,
+)
 
 
 class _NameBuilder:
@@ -105,24 +105,38 @@ def _spec_kind(spec: Any) -> str:
 
 def _transform_helper(
     flow_builder_state: _FlowBuilderState,
-    fn_spec: FunctionSpec,
+    fn_spec: FunctionSpec | Callable[..., Any],
     transform_args: list[tuple[Any, str | None]],
     name: str | None = None,
 ) -> DataSlice[Any]:
-    if not isinstance(fn_spec, FunctionSpec):
+    if isinstance(fn_spec, FunctionSpec):
+        kind = _spec_kind(fn_spec)
+        spec = fn_spec
+    elif callable(fn_spec) and (
+        op_kind := getattr(fn_spec, "__cocoindex_op_kind__", None)
+    ):
+        kind = op_kind
+        spec = op.EmptyFunctionSpec()
+    else:
         raise ValueError("transform() can only be called on a CocoIndex function")
 
-    return _create_data_slice(
-        flow_builder_state,
-        lambda target_scope, name: flow_builder_state.engine_flow_builder.transform(
-            _spec_kind(fn_spec),
-            dump_engine_object(fn_spec),
+    def _create_data_slice_inner(
+        target_scope: _engine.DataScopeRef | None, name: str | None
+    ) -> _engine.DataSlice:
+        result = flow_builder_state.engine_flow_builder.transform(
+            kind,
+            dump_engine_object(spec),
             transform_args,
             target_scope,
             flow_builder_state.field_name_builder.build_name(
                 name, prefix=_to_snake_case(_spec_kind(fn_spec)) + "_"
             ),
-        ),
+        )
+        return result
+
+    return _create_data_slice(
+        flow_builder_state,
+        _create_data_slice_inner,
         name,
     )
 
@@ -158,6 +172,7 @@ class _DataSliceState:
     def engine_data_slice(self) -> _engine.DataSlice:
         """
         Get the internal DataSlice.
+        This can be blocking.
         """
         if self._lazy_lock is None:
             if self._data_slice is None:
@@ -171,6 +186,13 @@ class _DataSliceState:
                     self._data_slice = self._data_slice_creator(None)
                 return self._data_slice
 
+    async def engine_data_slice_async(self) -> _engine.DataSlice:
+        """
+        Get the internal DataSlice.
+        This can be blocking.
+        """
+        return await asyncio.to_thread(lambda: self.engine_data_slice)
+
     def attach_to_scope(self, scope: _engine.DataScopeRef, field_name: str) -> None:
         """
         Attach the current data slice (if not yet attached) to the given scope.
@@ -245,7 +267,7 @@ class DataSlice(Generic[T]):
             f(scope)
 
     def transform(
-        self, fn_spec: op.FunctionSpec, *args: Any, **kwargs: Any
+        self, fn_spec: op.FunctionSpec | Callable[..., Any], *args: Any, **kwargs: Any
     ) -> DataSlice[Any]:
         """
         Apply a function to the data slice.
@@ -383,6 +405,7 @@ class DataCollector:
         /,
         *,
         primary_key_fields: Sequence[str],
+        attachments: Sequence[op.TargetAttachmentSpec] = (),
         vector_indexes: Sequence[index.VectorIndexDef] = (),
         vector_index: Sequence[tuple[str, index.VectorSimilarityMetric]] = (),
         setup_by_user: bool = False,
@@ -414,6 +437,10 @@ class DataCollector:
             target_name,
             _spec_kind(target_spec),
             dump_engine_object(target_spec),
+            [
+                {"kind": _spec_kind(att), **dump_engine_object(att)}
+                for att in attachments
+            ],
             dump_engine_object(index_options),
             self._engine_data_collector,
             setup_by_user,
@@ -432,7 +459,9 @@ class _FlowBuilderState:
     field_name_builder: _NameBuilder
 
     def __init__(self, full_name: str):
-        self.engine_flow_builder = _engine.FlowBuilder(full_name)
+        self.engine_flow_builder = _engine.FlowBuilder(
+            full_name, execution_context.event_loop
+        )
         self.field_name_builder = _NameBuilder()
 
     def get_data_slice(self, v: Any) -> _engine.DataSlice:
@@ -457,6 +486,7 @@ class _SourceRefreshOptions:
 class _ExecutionOptions:
     max_inflight_rows: int | None = None
     max_inflight_bytes: int | None = None
+    timeout: datetime.timedelta | None = None
 
 
 class FlowBuilder:
@@ -513,7 +543,7 @@ class FlowBuilder:
         )
 
     def transform(
-        self, fn_spec: FunctionSpec, *args: Any, **kwargs: Any
+        self, fn_spec: FunctionSpec | Callable[..., Any], *args: Any, **kwargs: Any
     ) -> DataSlice[Any]:
         """
         Apply a function to inputs, returning a DataSlice.
@@ -541,13 +571,19 @@ class FlowBuilder:
 class FlowLiveUpdaterOptions:
     """
     Options for live updating a flow.
+
+    - live_mode: Whether to perform live update for data sources with change capture mechanisms.
+    - reexport_targets: Whether to reexport to targets even if there's no change.
+    - print_stats: Whether to print stats during update.
     """
 
     live_mode: bool = True
+    reexport_targets: bool = False
     print_stats: bool = False
 
 
-class FlowUpdaterStatusUpdates(NamedTuple):
+@dataclass
+class FlowUpdaterStatusUpdates:
     """
     Status updates for a flow updater.
     """
@@ -667,28 +703,18 @@ class Flow:
     """
 
     _name: str
-    _full_name: str
-    _lazy_engine_flow: Callable[[], _engine.Flow] | None
+    _engine_flow_creator: Callable[[], _engine.Flow]
 
-    def __init__(
-        self, name: str, full_name: str, engine_flow_creator: Callable[[], _engine.Flow]
-    ):
+    _lazy_flow_lock: Lock
+    _lazy_query_handler_args: list[tuple[Any, ...]]
+    _lazy_engine_flow: _engine.Flow | None = None
+
+    def __init__(self, name: str, engine_flow_creator: Callable[[], _engine.Flow]):
         validate_flow_name(name)
-        validate_full_flow_name(full_name)
         self._name = name
-        self._full_name = full_name
-        engine_flow = None
-        lock = Lock()
-
-        def _lazy_engine_flow() -> _engine.Flow:
-            nonlocal engine_flow, lock
-            if engine_flow is None:
-                with lock:
-                    if engine_flow is None:
-                        engine_flow = engine_flow_creator()
-            return engine_flow
-
-        self._lazy_engine_flow = _lazy_engine_flow
+        self._engine_flow_creator = engine_flow_creator
+        self._lazy_flow_lock = Lock()
+        self._lazy_query_handler_args = []
 
     def _render_spec(self, verbose: bool = False) -> Tree:
         """
@@ -735,22 +761,27 @@ class Flow:
         """
         Get the full name of the flow.
         """
-        return self._full_name
+        return get_flow_full_name(self._name)
 
-    def update(self) -> _engine.IndexUpdateInfo:
+    def update(self, /, *, reexport_targets: bool = False) -> _engine.IndexUpdateInfo:
         """
         Update the index defined by the flow.
         Once the function returns, the index is fresh up to the moment when the function is called.
         """
-        return execution_context.run(self.update_async())
+        return execution_context.run(
+            self.update_async(reexport_targets=reexport_targets)
+        )
 
-    async def update_async(self) -> _engine.IndexUpdateInfo:
+    async def update_async(
+        self, /, *, reexport_targets: bool = False
+    ) -> _engine.IndexUpdateInfo:
         """
         Update the index defined by the flow.
         Once the function returns, the index is fresh up to the moment when the function is called.
         """
         async with FlowLiveUpdater(
-            self, FlowLiveUpdaterOptions(live_mode=False)
+            self,
+            FlowLiveUpdaterOptions(live_mode=False, reexport_targets=reexport_targets),
         ) as updater:
             await updater.wait_async()
         return updater.update_stats()
@@ -767,15 +798,33 @@ class Flow:
         """
         Get the engine flow.
         """
-        if self._lazy_engine_flow is None:
-            raise RuntimeError(f"Flow {self.full_name} is already removed")
-        return self._lazy_engine_flow()
+        if self._lazy_engine_flow is not None:
+            return self._lazy_engine_flow
+        return self._internal_flow()
 
     async def internal_flow_async(self) -> _engine.Flow:
         """
         Get the engine flow. The async version.
         """
-        return await asyncio.to_thread(self.internal_flow)
+        if self._lazy_engine_flow is not None:
+            return self._lazy_engine_flow
+        return await asyncio.to_thread(self._internal_flow)
+
+    def _internal_flow(self) -> _engine.Flow:
+        """
+        Get the engine flow. The async version.
+        """
+        with self._lazy_flow_lock:
+            if self._lazy_engine_flow is not None:
+                return self._lazy_engine_flow
+
+            engine_flow = self._engine_flow_creator()
+            self._lazy_engine_flow = engine_flow
+            for args in self._lazy_query_handler_args:
+                engine_flow.add_query_handler(*args)
+            self._lazy_query_handler_args = []
+
+            return engine_flow
 
     def setup(self, report_to_stdout: bool = False) -> None:
         """
@@ -787,9 +836,8 @@ class Flow:
         """
         Setup persistent backends of the flow. The async version.
         """
-        await make_setup_bundle([self]).describe_and_apply_async(
-            report_to_stdout=report_to_stdout
-        )
+        bundle = await make_setup_bundle_async([self])
+        await bundle.describe_and_apply_async(report_to_stdout=report_to_stdout)
 
     def drop(self, report_to_stdout: bool = False) -> None:
         """
@@ -806,9 +854,8 @@ class Flow:
         """
         Drop persistent backends of the flow. The async version.
         """
-        await make_drop_bundle([self]).describe_and_apply_async(
-            report_to_stdout=report_to_stdout
-        )
+        bundle = await make_drop_bundle_async([self])
+        await bundle.describe_and_apply_async(report_to_stdout=report_to_stdout)
 
     def close(self) -> None:
         """
@@ -822,6 +869,53 @@ class Flow:
         with _flows_lock:
             del _flows[self.name]
 
+    def add_query_handler(
+        self,
+        name: str,
+        handler: Callable[[str], Any],
+        /,
+        *,
+        result_fields: QueryHandlerResultFields | None = None,
+    ) -> None:
+        """
+        Add a query handler to the flow.
+        """
+        async_handler = to_async_call(handler)
+
+        async def _handler(query: str) -> dict[str, Any]:
+            handler_result = await async_handler(query)
+            return {
+                "results": [
+                    [(k, dump_engine_object(v)) for (k, v) in result.items()]
+                    for result in handler_result.results
+                ],
+                "query_info": dump_engine_object(handler_result.query_info),
+            }
+
+        handler_info = dump_engine_object(QueryHandlerInfo(result_fields=result_fields))
+        with self._lazy_flow_lock:
+            if self._lazy_engine_flow is not None:
+                self._lazy_engine_flow.add_query_handler(name, _handler, handler_info)
+            else:
+                self._lazy_query_handler_args.append((name, _handler, handler_info))
+
+    def query_handler(
+        self,
+        name: str | None = None,
+        result_fields: QueryHandlerResultFields | None = None,
+    ) -> Callable[[Callable[[str], Any]], Callable[[str], Any]]:
+        """
+        A decorator to declare a query handler.
+        """
+
+        def _inner(handler: Callable[[str], Any]) -> Callable[[str], Any]:
+            self.add_query_handler(
+                name or handler.__qualname__, handler, result_fields=result_fields
+            )
+            return handler
+
+        return _inner
+
 
 def _create_lazy_flow(
     name: str | None, fl_def: Callable[[FlowBuilder, DataScope], None]
@@ -831,19 +925,18 @@ def _create_lazy_flow(
     The flow will be built the first time when it's really needed.
     """
     flow_name = _flow_name_builder.build_name(name, prefix="_flow_")
-    flow_full_name = get_flow_full_name(flow_name)
 
     def _create_engine_flow() -> _engine.Flow:
+        flow_full_name = get_flow_full_name(flow_name)
+        validate_full_flow_name(flow_full_name)
         flow_builder_state = _FlowBuilderState(flow_full_name)
         root_scope = DataScope(
             flow_builder_state, flow_builder_state.engine_flow_builder.root_scope()
         )
         fl_def(FlowBuilder(flow_builder_state), root_scope)
-        return flow_builder_state.engine_flow_builder.build_flow(
-            execution_context.event_loop
-        )
+        return flow_builder_state.engine_flow_builder.build_flow()
 
-    return Flow(flow_name, flow_full_name, _create_engine_flow)
+    return Flow(flow_name, _create_engine_flow)
 
 
 _flows_lock = Lock()
@@ -973,11 +1066,19 @@ def _get_data_slice_annotation_type(
 _transform_flow_name_builder = _NameBuilder()
 
 
-class TransformFlowInfo(NamedTuple):
+@dataclass
+class TransformFlowInfo(Generic[T]):
     engine_flow: _engine.TransientFlow
     result_decoder: Callable[[Any], T]
 
 
+@dataclass
+class FlowArgInfo:
+    name: str
+    type_hint: Any
+    encoder: Callable[[Any], Any]
+
+
 class TransformFlow(Generic[T]):
     """
     A transient transformation flow that transforms in-memory data.
@@ -985,16 +1086,14 @@ class TransformFlow(Generic[T]):
 
     _flow_fn: Callable[..., DataSlice[T]]
     _flow_name: str
-    _flow_arg_types: list[Any]
-    _param_names: list[str]
+    _args_info: list[FlowArgInfo]
 
     _lazy_lock: asyncio.Lock
-    _lazy_flow_info: TransformFlowInfo | None = None
+    _lazy_flow_info: TransformFlowInfo[T] | None = None
 
     def __init__(
         self,
         flow_fn: Callable[..., DataSlice[T]],
-        flow_arg_types: Sequence[Any],
         /,
         name: str | None = None,
     ):
@@ -1002,39 +1101,11 @@ class TransformFlow(Generic[T]):
         self._flow_name = _transform_flow_name_builder.build_name(
             name, prefix="_transform_flow_"
         )
-        self._flow_arg_types = list(flow_arg_types)
         self._lazy_lock = asyncio.Lock()
 
-    def __call__(self, *args: Any, **kwargs: Any) -> DataSlice[T]:
-        return self._flow_fn(*args, **kwargs)
-
-    @property
-    def _flow_info(self) -> TransformFlowInfo:
-        if self._lazy_flow_info is not None:
-            return self._lazy_flow_info
-        return execution_context.run(self._flow_info_async())
-
-    async def _flow_info_async(self) -> TransformFlowInfo:
-        if self._lazy_flow_info is not None:
-            return self._lazy_flow_info
-        async with self._lazy_lock:
-            if self._lazy_flow_info is None:
-                self._lazy_flow_info = await self._build_flow_info_async()
-            return self._lazy_flow_info
-
-    async def _build_flow_info_async(self) -> TransformFlowInfo:
-        flow_builder_state = _FlowBuilderState(self._flow_name)
-        sig = inspect.signature(self._flow_fn)
-        if len(sig.parameters) != len(self._flow_arg_types):
-            raise ValueError(
-                f"Number of parameters in the flow function ({len(sig.parameters)}) "
-                f"does not match the number of argument types ({len(self._flow_arg_types)})"
-            )
-
-        kwargs: dict[str, DataSlice[T]] = {}
-        for (param_name, param), param_type in zip(
-            sig.parameters.items(), self._flow_arg_types
-        ):
+        sig = inspect.signature(flow_fn)
+        args_info = []
+        for param_name, param in sig.parameters.items():
             if param.kind not in (
                 inspect.Parameter.POSITIONAL_OR_KEYWORD,
                 inspect.Parameter.KEYWORD_ONLY,
@@ -1042,35 +1113,68 @@ class TransformFlow(Generic[T]):
                 raise ValueError(
                     f"Parameter `{param_name}` is not a parameter can be passed by name"
                 )
-            encoded_type = encode_enriched_type(param_type)
-            if encoded_type is None:
-                raise ValueError(f"Parameter `{param_name}` has no type annotation")
-            engine_ds = flow_builder_state.engine_flow_builder.add_direct_input(
-                param_name, encoded_type
+            value_type_annotation: type | None = _get_data_slice_annotation_type(
+                param.annotation
             )
-            kwargs[param_name] = DataSlice(
+            if value_type_annotation is None:
+                raise ValueError(
+                    f"Parameter `{param_name}` for {flow_fn} has no value type annotation. "
+                    "Please use `cocoindex.DataSlice[T]` where T is the type of the value."
+                )
+            encoder = make_engine_value_encoder(
+                analyze_type_info(value_type_annotation)
+            )
+            args_info.append(FlowArgInfo(param_name, value_type_annotation, encoder))
+        self._args_info = args_info
+
+    def __call__(self, *args: Any, **kwargs: Any) -> DataSlice[T]:
+        return self._flow_fn(*args, **kwargs)
+
+    @property
+    def _flow_info(self) -> TransformFlowInfo[T]:
+        if self._lazy_flow_info is not None:
+            return self._lazy_flow_info
+        return execution_context.run(self._flow_info_async())
+
+    async def _flow_info_async(self) -> TransformFlowInfo[T]:
+        if self._lazy_flow_info is not None:
+            return self._lazy_flow_info
+        async with self._lazy_lock:
+            if self._lazy_flow_info is None:
+                self._lazy_flow_info = await self._build_flow_info_async()
+            return self._lazy_flow_info
+
+    async def _build_flow_info_async(self) -> TransformFlowInfo[T]:
+        flow_builder_state = _FlowBuilderState(self._flow_name)
+        kwargs: dict[str, DataSlice[T]] = {}
+        for arg_info in self._args_info:
+            encoded_type = encode_enriched_type(arg_info.type_hint)
+            if encoded_type is None:
+                raise ValueError(f"Parameter `{arg_info.name}` has no type annotation")
+            engine_ds = flow_builder_state.engine_flow_builder.add_direct_input(
+                arg_info.name, encoded_type
+            )
+            kwargs[arg_info.name] = DataSlice(
                 _DataSliceState(flow_builder_state, engine_ds)
             )
 
-        output = self._flow_fn(**kwargs)
-        flow_builder_state.engine_flow_builder.set_direct_output(
-            _data_slice_state(output).engine_data_slice
-        )
+        output = await asyncio.to_thread(lambda: self._flow_fn(**kwargs))
+        output_data_slice = await _data_slice_state(output).engine_data_slice_async()
+
+        flow_builder_state.engine_flow_builder.set_direct_output(output_data_slice)
         engine_flow = (
             await flow_builder_state.engine_flow_builder.build_transient_flow_async(
                 execution_context.event_loop
             )
         )
-        self._param_names = list(sig.parameters.keys())
-
-        engine_return_type = (
-            _data_slice_state(output).engine_data_slice.data_type().schema()
-        )
+        engine_return_type = output_data_slice.data_type().schema()
         python_return_type: type[T] | None = _get_data_slice_annotation_type(
-            sig.return_annotation
+            inspect.signature(self._flow_fn).return_annotation
         )
         result_decoder = make_engine_value_decoder(
-            [], engine_return_type["type"], analyze_type_info(python_return_type)
+            [],
+            decode_engine_value_type(engine_return_type["type"]),
+            analyze_type_info(python_return_type),
         )
 
         return TransformFlowInfo(engine_flow, result_decoder)
@@ -1099,13 +1203,14 @@ class TransformFlow(Generic[T]):
         """
         flow_info = await self._flow_info_async()
         params = []
-        for i, arg in enumerate(self._param_names):
+        for i, arg_info in enumerate(self._args_info):
             if i < len(args):
-                params.append(encode_engine_value(args[i]))
+                arg = args[i]
             elif arg in kwargs:
-                params.append(encode_engine_value(kwargs[arg]))
+                arg = kwargs[arg]
             else:
                 raise ValueError(f"Parameter {arg} is not provided")
+            params.append(arg_info.encoder(arg))
         engine_result = await flow_info.engine_flow.evaluate_async(params)
         return flow_info.result_decoder(engine_result)
 
@@ -1116,53 +1221,47 @@ def transform_flow() -> Callable[[Callable[..., DataSlice[T]]], TransformFlow[T]
     """
 
     def _transform_flow_wrapper(fn: Callable[..., DataSlice[T]]) -> TransformFlow[T]:
-        sig = inspect.signature(fn)
-        arg_types = []
-        for param_name, param in sig.parameters.items():
-            if param.kind not in (
-                inspect.Parameter.POSITIONAL_OR_KEYWORD,
-                inspect.Parameter.KEYWORD_ONLY,
-            ):
-                raise ValueError(
-                    f"Parameter `{param_name}` is not a parameter can be passed by name"
-                )
-            value_type_annotation: type[T] | None = _get_data_slice_annotation_type(
-                param.annotation
-            )
-            if value_type_annotation is None:
-                raise ValueError(
-                    f"Parameter `{param_name}` for {fn} has no value type annotation. "
-                    "Please use `cocoindex.DataSlice[T]` where T is the type of the value."
-                )
-            arg_types.append(value_type_annotation)
-
-        _transform_flow = TransformFlow(fn, arg_types)
+        _transform_flow = TransformFlow(fn)
         functools.update_wrapper(_transform_flow, fn)
         return _transform_flow
 
     return _transform_flow_wrapper
 
 
-def make_setup_bundle(flow_iter: Iterable[Flow]) -> SetupChangeBundle:
+async def make_setup_bundle_async(flow_iter: Iterable[Flow]) -> SetupChangeBundle:
     """
     Make a bundle to setup flows with the given names.
     """
     full_names = []
     for fl in flow_iter:
-        fl.internal_flow()
+        await fl.internal_flow_async()
         full_names.append(fl.full_name)
     return SetupChangeBundle(_engine.make_setup_bundle(full_names))
 
 
+def make_setup_bundle(flow_iter: Iterable[Flow]) -> SetupChangeBundle:
+    """
+    Make a bundle to setup flows with the given names.
+    """
+    return execution_context.run(make_setup_bundle_async(flow_iter))
+
+
+async def make_drop_bundle_async(flow_iter: Iterable[Flow]) -> SetupChangeBundle:
+    """
+    Make a bundle to drop flows with the given names.
+    """
+    full_names = []
+    for fl in flow_iter:
+        await fl.internal_flow_async()
+        full_names.append(fl.full_name)
+    return SetupChangeBundle(_engine.make_drop_bundle(full_names))
+
+
 def make_drop_bundle(flow_iter: Iterable[Flow]) -> SetupChangeBundle:
     """
     Make a bundle to drop flows with the given names.
     """
-    full_names = []
-    for fl in flow_iter:
-        fl.internal_flow()
-        full_names.append(fl.full_name)
-    return SetupChangeBundle(_engine.make_drop_bundle(full_names))
+    return execution_context.run(make_drop_bundle_async(flow_iter))
 
 
 def setup_all_flows(report_to_stdout: bool = False) -> None:
