@@ -14,6 +14,7 @@ class LlmApiType(Enum):
     OPEN_ROUTER = "OpenRouter"
     VOYAGE = "Voyage"
     VLLM = "Vllm"
+    BEDROCK = "Bedrock"
 
 
 @dataclass
@@ -26,6 +27,16 @@ class VertexAiConfig:
     region: str | None = None
 
 
+@dataclass
+class OpenAiConfig:
+    """A specification for a OpenAI LLM."""
+
+    kind = "OpenAi"
+
+    org_id: str | None = None
+    project_id: str | None = None
+
+
 @dataclass
 class LlmSpec:
     """A specification for a LLM."""
@@ -33,4 +44,4 @@ class LlmSpec:
     api_type: LlmApiType
     model: str
     address: str | None = None
-    api_config: VertexAiConfig | None = None
+    api_config: VertexAiConfig | OpenAiConfig | None = None
